/* -*-c++-*-*/
// COPYRIGHT DASSAULT SYSTEMES 1997
//=============================================================================
//
// CATDMUsrMethod:
// Contains all the user exits related to BOM
//
//=============================================================================
// Usage Notes:
//
//=============================================================================
// Sep. 97   Creation                                  
//=============================================================================

#ifndef CATDMUsrMethod_h
#define CATDMUsrMethod_h

#define ExportedByLV0XUSR

#include <time.h>

#include <CATCollec.h>
class CATVPMNode;
class CATLISTP(CATVPMNode);
class SdaiAppInstance;
class CATDMUsrInstance;

    /**
     * Modification types for commit.
     */
enum ModificationType {
    ADDLINK=1,
    REMOVELINK=2,
    MOVELINK=4,
    ADDMOVELINK=4+1,
    REPLACELINK=8,
    MOVEREPLACELINK=4+8,
    RENAMELINK=16,
    ADDRENAMELINK=16+1,
    ADDLINKTOCONF=32+1,
    MOVESTEPLINK=64
    };

    /**
     * Action types for versioning.
     */
enum RevVerAction
{
    ACTION_NEW,
    ACTION_UPDATE
};

    /**
     * Actions for config-handlers control.
     */
enum CHAction
{
    CH_CREATE,           // Create a config-handler
    CH_RENAME,           // Rename a config-handler
    CH_DUPLICATE,        // Duplicate a config-handler
    CH_DELETE,           // Delete a config-handler
    CH_EDIT_CONTENT,     // Edit properties of a config-handler
    CH_EDIT_LINK,        // Edit child/brother links of a config-handler
    CH_CHANGE_OWNER      // Change ownership of a config-handler
};

    /**
     * Actions for option-dictionary control.
     */
enum DicoAction
{
    OPT_CREATE,          // Create an option in the dictionary
    OPT_RENAME,          // Rename an option in the dictionary
    OPT_DELETE,          // Delete an option in the dictionary
    OPT_EDIT,            // Edit properties of an option in the dictionary
    CATEG_CREATE,        // Create a category in the dictionary
    CATEG_RENAME,        // Rename a category in the dictionary
    CATEG_DELETE,        // Delete a category in the dictionary
    CATEG_EDIT,          // Edit properties of a category in the dictionary
    CATEG_CHANGE_OWNER,  // Change ownership of a category in the dictionary
    OPT_WHERE_USED,      // Search where used the option in rules of the dictionary
    CATEG_WHERE_USED,    // Search where used the category in rules of the dictionary
    DICO_UPDATE_ASS,     // Update the dictionary from the associated assembly
    DICO_IMPORT_CORP,    // Import categories from corporate to product dictionary
    DICO_DELETE          // Delete the product dictionary
};

    /**
     * Actions for rules control in option-dictionary.
     */
enum DicoRulesAction
{
    RULE_CREATE,         // Create a directive rule in the dictionary
    RULE_RENAME,         // Rename a directive rule in the dictionary
    RULE_DELETE,         // Delete a directive rule in the dictionary
    RULE_EDIT,           // Edit properties of a directive rule in the dictionary
    RULE_CHANGE_OWNER,   // Change ownership of a directive rule in the dictionary
    RULE_IMPORT_CORP,    // Import directive rules from corporate to product dictionary
    PACKAGE_CREATE,      // Create a package in the dictionary
    PACKAGE_RENAME,      // Rename a package in the dictionary
    PACKAGE_DELETE,      // Delete a package in the dictionary
    PACKAGE_EDIT,        // Edit properties of a package in the dictionary
    PACKAGE_CHANGE_OWNER,// Change ownership of a package in the dictionary
    PACKAGE_IMPORT_CORP, // Import packages from corporate to product dictionary
    MATRIX_CREATE,       // Create a matrix in the dictionary
    MATRIX_RENAME,       // Rename a matrix in the dictionary
    MATRIX_DELETE,       // Delete a matrix in the dictionary
    MATRIX_EDIT,         // Edit properties of a matrix in the dictionary
    MATRIX_EDIT_ROW,     // Modify a row of a matrix (options combination) in the dictionary
    MATRIX_CHANGE_OWNER, // Change ownership of a matrix in the dictionary
    MATRIX_IMPORT_CORP   // Import matrices from corporate to product dictionary
};

    /**
     * Actions for milestones control.
     */
enum MilestonesAction
{
    PROGRAM_CREATE,       // Create a program for product
    PROGRAM_RENAME,       // Rename a program for product
    MILESTONE_CREATE,     // Create a milestone
    MILESTONE_RENAME,     // Rename a milestone
    MILESTONE_DELETE,     // Delete a milestone
    MILESTONE_EDIT,       // Edit properties of a milestone 
    MILESTONE_FREEZE,     // Freeze a milestone
    MILESTONE_UNFREEZE,   // Unfreeze a milestone
    ALL_MILESTONES_DELETE // Delete all programs/milestones for product
};

// Config-handlers visibility
#define CH_PUBLIC            1
#define CH_PRIVATE          (1 << 1)
#define CH_DATAGROUP        (1 << 2)

// Modes to commit modification
#define COMMIT_ACTION        1
#define COMMIT_CFG_PARAM    (1 << 1)
#define COMMIT_EXISTING_EFF (1 << 2)
#define COMMIT_STEP_FILE    (1 << 3)


#ifndef PartAsmID_def
#define PartAsmID_def
typedef struct PartAsmId_ {
    char   coid[8];
    char   compid[8];
    char   caenv[8];
    char   catab[18];
    char   clien[8];    // link COMPID (towards father)
    double xmat[12];    // transformation matrix 
    int	   level;
} PartAsmID;
#endif

#ifndef ObjID_def
#define ObjID_def
typedef struct  
{
    char   Xcoid[17];   // Coid (hexadecimal notation)
    char   Xcompid[17]; // Compid (hexadecimal notation) 
    double coid;        // Coid (binary)
    double compid;      // Compid (binary)
    char   caenv[9];
    char   catab[19];
} ObjID;
#endif

#ifndef EffParam_def
#define EffParam_def
typedef struct  
{
    char   Type[40];           // Type of the effectivity:
			       //   modif, date, range, option_undef, pack_undef,
			       //   option_date, option_range, pack_date, pack_range
    char   Modif[40];          // Modification name (Type = modif)
    char   Option[40];         // Option name (Type = option_XXX)
    int    NbOpt;              // Nb options in OptList (Type = pack_XXX)
    char** OptList;            // Options (combined with AND operator) (Type = pack_XXX)
    int    BotRange;           // Bottom value for range (Type = range, XXX_range)
    int    TopRange;           // Top value for range (Type = range, XXX_range)
    time_t BegDate;            // Start value for date (Type = date, XXX_date)
    time_t EndDate;            // End value for date (Type = date, XXX_date)
    int    Sign;               // Sign of the effectivity (1 or -1 with -1 = NOT)
    char   StartMilestProg[40];// Program name for start milestone (range or date)
    char   StartMilestId[40];  // Milestone name for start milestone (range or date)
    char   EndMilestProg[40];  // Program name for end milestone (range or date)
    char   EndMilestId[40];    // Milestone name for end milestone (range or date)
} EffParam;
#endif

#ifndef Error_def
#define Error_def
typedef struct  
{
    int   Ier;         // Error code
    char* ShortMsg;    // Short message (deleted by VPM if not NULL and if Ier!=0)
    char* LongMsg;     // Long message (deleted by VPM if not NULL and if Ier!=0)
} Error;
#endif

/**
 * Note:<BR>
 *   
 * Following convention is used for naming parameters: parameter names prefixed by:
 * <dl>
 *   <dd> <tt>i</tt> are input parameters
 *   <dd> <tt>o</tt> are ouput parameters
 *   <dd> <tt>io</tt> are input/output parameters
 * </dl>
 */


class ExportedByLV0XUSR CATDMUsrMethod
{
  public:
    //----------------------------------------
    // Additional tools and methods management
    //----------------------------------------
    static Error ExecuteUsrMethod(const char *iMethodName,
				  const int &iNbSelectedInstances,
				  ObjID *iSelectedObjects);

    static Error OpenPSNWithUsrMethod(const char* iMethodName,
				      const int& iNbSelectedInstances,
				      int* iDepthArray,
				      PartAsmID** iObjArray,
				      double** iComputedXmatArray);
   

    //-------------------
    // PSN save management
    //-------------------
    /**
     * Allows customization of the proposed values list in the Modification Name combo, when
     * commiting changes on a configurable part in PSN.
     * <p>
     * If method does not provide a history list (<tt>oHistSize=0</tt>), the caller method will
     * use default history.<BR>
     * Allocation of <tt>oHistory</tt>'s elements has to be done in function
     * GetConfNameHistory.<BR>
     * Desallocation is the responsability of VPM.
     * <p>
     * @param iUser
     *    user name
     * @param iOrg
     *    user organization
     * @param oHistory
     *    points on the list of proposed values
     * @param oHistSize
     *    number of proposed values
     */
    static void GetConfNameHistory(const char iUser[9], const char iOrg[11],
				   char*** oHistory, int* oHistSize);

    /**
     * Allows user to validate keyed flag versus current modifications. 
     * <p>
     * @param iModificationFlag
     *    flag which is going to be used in order to save following modifications
     * @param iNbModifications
     *    number of modifications 
     * @param iTypeOfModifications
     *    list of types for each modification (with iNbModifications rows)
     * @param iAddedLinks
     *    list of added links (with iNbModifications rows)
     * @param iRemovedLinks
     *    list of removed links (with iNbModifications rows)
     * <p>
     * @return
     *     If method return an error, another modification flag must be keyed or<BR>
     *     commit canceled.
     */
    static Error IsAValidFlagForCommit(const char* iModificationFlag,
				       const int iNbModifications,
				       int* iTypeOfModifications,
				       ObjID* iAddedLinks,
				       ObjID* iRemovedLinks);

    /**
     * Allows user to validate keyed parameters versus current modifications. 
     * <p>
     * @param iNbParam
     *    number of rows in the parameters matrix
     * @param iParamPtr
     *    list of parameters (with iNbParam rows)
     * @param iNbModifications
     *    number of modifications 
     * @param iTypeOfModifications
     *    list of types for each modification (with iNbModifications rows)
     * @param iAddedLinks
     *    list of added links (with iNbModifications rows)
     * @param iRemovedLinks
     *    list of removed links (with iNbModifications rows)
     * <p>
     * @return
     *     If method return an error, parameters must be changed or commit canceled.
     */
    static Error IsAValidParamListForCommit(const int iNbParam, EffParam *iParamPtr,
					    const int iNbModifications,
					    int* iTypeOfModifications,
					    ObjID* iAddedLinks, ObjID* iRemovedLinks);

    /**
     * Allows user to validate the commit of current modifications with existing
     * effectivity parameters. 
     * <p>
     * @param iNbModifications
     *    number of modifications 
     * @param iTypeOfModifications
     *    list of types for each modification (with iNbModifications rows)
     * @param iAddedLinks
     *    list of added links (with iNbModifications rows)
     * @param iRemovedLinks
     *    list of removed links (with iNbModifications rows)
     * <p>
     * @return
     *     If method return an error, commit canceled.
     */
    static Error IsExistingParamValidForCommit(const int iNbModifications,
					       int* iTypeOfModifications,
					       ObjID* iAddedLinks, ObjID* iRemovedLinks);
    
    /**
     * User exit called when saving Product Structure modifications.
     * <p>
     * @return
     *   <tt>1</tt> Action flow will test whether the current user/role has the right
     *              to write the modifications (maintenance phase)<BR>
     *   <tt>0</tt> Action flow won't be called when saving the modifications
     *              (development phase)
     */ 
    static int ActionFlowCheck(const char *iModificationFlag, 
			       const char *iUser);
    
    static Error SuccessfullyCommitInGraph(const char* iModificationFlag,
					   const int iNbModifications,
					   int* iTypeOfModifications,
					   ObjID* iAddedLinks,
					   ObjID* iRemovedLinks);
    
    static void SuccessfullyRollBackInGraph();

    
    //-------------------------
    // Modifications management
    //-------------------------

    /**
     * User exit called when validating action and commiting PSN graph modifications.
     * <p>
     * @param iModif 
     *    Modification to validate
     * @param oType 
     *    Type of the validation (option, range, date, pack, option_range, option_date,
     *    pack_range, pack_date)
     * @param oOptListSize
     *    Number of options in oOptList<BR>
     *    = 1 if type is equal "option", "option_range" or "option_date"<BR>
     *    = n if type is equal "pack", "pack_range" or "pack_date"
     * @param oOptList
     *    significant only if oType is equal to "option", "option_range",
     *    "option_date", "pack", "pack_range" or "pack_date"
     * @param oRangeBot
     *    significant only if oType is equal to "range", "option_range"
     *    or "pack_range"
     * @param oRangeTop
     *    significant only if oType is equal to "range", "option_range"
     *    or "pack_range"
     * @param oDateBeg
     *    significant only if oType is equal to "date", "option_date"
     *    or "pack_date"
     * @param oDateEnd
     *    significant only if oType is equal to "date", "option_date"
     *    or "pack_date"
     * <p>
     * @return
     *    If Error.Ier equals 0 validation is done, else error message is displayed
     */
    static Error DefaultModValidation(const char *iModif, char **oType, int &oOptListSize, char ***oOptList, int &oRangeBot, int &oRangeTop, time_t &oDateBeg, time_t &oDateEnd);

    /**
     * User exit called when validating a modification.
     * <p>
     * @param iModification 
     *    Modification to validate
     * @param iUser
     *    Identifier of the user
     * @param iValidationType 
     *    Type of the validation (option, range, date, pack, option_range, option_date,
     *    pack_range, pack_date)
     * @param iNbOptions
     *    Number of options in iOptionList<BR>
     *    = 1 if type is equal "option", "option_range" or "option_date"<BR>
     *    = n if type is equal "pack", "pack_range" or "pack_date"
     * @param iOptionList
     *    significant only if iValidationType is equal to "option", "option_range",
     *    "option_date", "pack", "pack_range" or "pack_date"
     * @param iRangeBot
     *    significant only if iValidationType is equal to "range", "option_range"
     *    or "pack_range"
     * @param iRangeTop
     *    significant only if iValidationType is equal to "range", "option_range"
     *    or "pack_range"
     * @param iDateBeg
     *    significant only if iValidationType is equal to "date", "option_date"
     *    or "pack_date"
     * @param iDateEnd
     *    significant only if iValidationType is equal to "date", "option_date"
     *    or "pack_date"
     * <p>
     * @return
     *    If Error.Ier equals 0 validation is allowed, else error message is displayed
     */
    static Error AuthorizedModValidation(const char *iModification,
					 const char *iUser,                               
					 const char *iValidationType,
					 const int iNbOptions,
					 char **iOptionList,
					 const int iRangeBot,   const int iRangeTop,
					 const time_t iDateBeg, const time_t iDateEnd);
  
    /**
     * User exit called when promoting a modification.
     * <p>
     * @return
     *    If Error.Ier equals 0 promote is allowed, else error message is displayed
     */
    static Error AuthorizedModPromote(const char *iModification, const char *iUser);                                
    /**
     * User exit called when demoting a modification.
     * <p>
     * @return
     *    If Error.Ier equals 0 demote is allowed, else error message is displayed
     */
    static Error AuthorizedModDemote(const char *iModification, const char *iUser);                                
    /**
     * User exit called when deleting a modification.
     * <p>
     * @return
     *    If Error.Ier equals 0 delete is allowed, else error message is displayed
     */
    static Error AuthorizedModDelete(const char *iModification, const char *iUser);                                
    /**
     * User exit called when creating a modification.
     * <p>
     * @return
     *    <tt>0</tt> iModification will be of Personal type and belongs to iUser 
     *               (only iUser will see iModification until promoted)<BR>
     *    <tt>1</tt> iModification will be of Public type (everybody can see it)
     */
    static int ModCreateOwner(const char *iModification, const char *iUser);  

    
    //-----------------------------------------------------------------
    // Check user authorization for basic operations
    // Operations are performed if no error (error code =0) is returned
    //-----------------------------------------------------------------
    // Copy Instance
    static Error AuthorizedCopyInstance(const int & NbSelectedInstances, ObjID * iSelectedInstances);

    // Cut link
    static Error AuthorizedCut(const ObjID& iLink);

    // Cut Instance
    static Error AuthorizedCutInstance(const int & NbSelectedInstances, ObjID * iSelectedInstances);

    // Move Link 
    static Error AuthorizedMove(const ObjID& iLink);

    // Create Link (with paste method)
    static Error AuthorizedPaste(const ObjID& iFather, const ObjID& iChild);

    // Paste Instance on existing part
    static Error AuthorizedPasteInstance(const int & NbSelectedInstances, ObjID * iSelectedInstances);
    
    // CheckIn object
    static Error AuthorizedCheckIn(const ObjID& iObj);

    // CheckOut object
    static Error AuthorizedCheckOut(const ObjID& iObj);

    // Delete object
    static Error AuthorizedDelete(const ObjID& iObj);

    // Create object
    static Error AuthorizedCreate(const ObjID& iObj);

    // Define Instance
    static Error AuthorizedDefineInstance(const int & NbSelectedInstances, ObjID * iObj);

    // Edit Instance
    static Error AuthorizedEditInstance(const int & NbSelectedInstances, ObjID * iObj);

    // Update object
    static Error AuthorizedUpdate(const ObjID& iObj);
  
    // Replace 
    static Error AuthorizedReplace(const ObjID& iFather,const ObjID& iLink,
				   const ObjID& iChild,const ObjID& iNewChild);

    // Query On Instance
    static Error AuthorizedQueryOnInstance(const ObjID& iInstance);

    /**
     * Checks autorisation before saving model in CATIA V4.
     * <p>
     * @return 
     *   <tt>=&nbsp;0</tt> means save is authorized in CATIA V4<BR>
     *   <tt><>0</tt> it is an error code => save in CATIA will not be performed
     *                    
     */
    static int AuthorizedCATIASave(const ObjID& iObj);

    /**
     * Checks autorisation before saving document in CATIA V5.
     * <p>
     * @return 
     *   <tt>=&nbsp;0</tt> means save is authorized in CATIA V5<BR>
     *   <tt><>0</tt> it is an error code => save in CATIA will not be performed
     */
    static Error AuthorizedCATIAV5Save(const ObjID& iObj);


    // Change maturity
    static Error AuthorizedPromotion(const ObjID& iObj,
				     const char* iOldStatus,
				     const char* iNewStatus);
    
    /**
     * User exit called when changing ownership.
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error AuthorizedReassign(const ObjID& iObj,
				    const char iUser[9], 
				    const char iOrg[11],     
				    const char iSite[11]);     


    /**
     * User exit called when duplicating representations from one part to another.
     * <p>
     * @param ioNbObj
     *   <tt>Input&nbsp;</tt> Number of candidate representations.<BR>
     *   <tt>Output</tt> Number of representations to duplicate.
     * @param ioObjs
     *   <tt>Input&nbsp;</tt> Candidate representations.<BR>
     *   <tt>Output</tt> Representations to duplicate.
     * <p>
     * @return
     *    If Error.Ier equals 0 duplication is allowed, else error message is displayed.
     */
    static Error AuthorizedDuplicatedRepresentations(int*    ioNbObj,
                                                     ObjID** ioObjs );


    /**
     * User exit called when duplicating asssembly links from one part to another.
     * <p>
     * @param ioNbObj
     *   <tt>Input&nbsp;</tt> Number of candidate assembly links.<BR>
     *   <tt>Output</tt> Number of asssembly links to duplicate.
     * @param ioObjs
     *   <tt>Input&nbsp;</tt> Candidate assembly links.<BR>
     *   <tt>Output</tt> Assembly link to duplicate.
     * <p>
     * @return
     *    If Error.Ier equals 0 duplication is allowed, else error message is displayed.
     */
    static Error AuthorizedDuplicatedAssemblyLinks(int*    ioNbObj,
						   ObjID** ioObjs );

    //----------------------
    // Attributes management 
    //----------------------
    /**
     * User exit called when user create new panel (insert, query or update) if
     * $AUTHORIZEDVALUES or $HELPVALUES exist but empty.
     * <p>
     * <b>This user-exit is called several times. At each call, you only have to  
     * fill the  output parameters which are not NULL. </b>
     * <p>
     * @param oAuthorized
     *   <tt>-1</tt> Help values without default value<BR>
     *   <tt>-2</tt> Help values with first value as default value<BR>
     *   <tt>&nbsp;1</tt> Authorized values without default value<BR>
     *   <tt>&nbsp;2</tt> Authorized values with first value as default value
     *
     * @param oNbValues
     *   number of values returned
     * @param oValues
     *   list of values (this array is allocated and deleted by VPM)
     */
    static void GetInputListOfValues(const char* iCatab,
				     const char* iAttribute, 
				     int* oAuthorized=0,
				     int* oNbValues=0,
				     char*** oValues=0);
    
    /**
     * Allows to check attribute values or to valuate company attributes before
     * writing a part or an object.
     * <p>
     * This method is called before checking if an object with the input values
     * already exists, and, thus before ComputeValuesBeforeWrite().<BR>
     * As opposed to the latter, this method shall only modify external attributes,
     * that is, all but those prefixed by "$" and "C_" (except for C_NICKNAME and
     * C_LAST_REPOSITORY for representations).<BR>
     * For example, it is possible to hide an "S_" (but mandatory) attribute from
     * the user in the insert panel, and fill it here.
     * <p>
     * @param iUpdate
     *   <tt>0</tt>  user-exit called in write context<BR>
     *   <tt>1</tt>  user-exit called in update context
     * @param iCaenv
     *   Owner of the environment into which the object is being written
     * @param iCatab
     *   Name of the table into which the object is being written
     * @param iNbAttPart
     *   Number of attributes given for the part
     * @param iNbAttObj  
     *   Number of attributes given for the object
     * @param ioAttNames   
     *   Names of the given attributes for the part and the object being written
     *   (array size: iNbAttPart + iNbAttObj)
     * @param ioAttVals
     *   Values of the given attributes for the part and the object being written
     *   (array size: iNbAttPart + iNbAttObj)
     * <p>
     * @return
     * The method must return 0 if the operation is accepted or error code if it is 
     * refused. In such a case, an error message is displayed in an error notification
     * panel and the operation is canceled.<BR>
     * The attributes given in the list are the ones that were valuated in the insert
     * panel. If you wish to valuate a "hidden" attribute, you will have to reallocate
     * AttNames and AttVals and add the attribute and its value.
     */
    static Error ComputeUserValuesBeforeWrite(const int iUpdate,
					      const char* const iCaenv, 
					      const char* const iCatab,
					      int *iNbAttPart, 
					      int *iNbAttObj,
					      char ***ioAttNames, 
					      char ***ioAttVals);
    
    /**
     * Allows to check attribute values or to valuate company attributes before writing
     * a part or an object.
     * <p>
     * @param iUpdate
     *   <tt>0</tt>  user-exit called in write context<BR>
     *   <tt>1</tt>  user-exit called in update context
     * @param iCaenv
     *   Owner of the environment into which the object is being written
     * @param iCatab
     *   Name of the table into which the object is being written
     * @param iNbAttPart
     *   Number of attributes given for the part
     * @param iNbAttObj  
     *   Number of attributes given for the object
     * @param ioAttNames   
     *   Names of the given attributes for the part and the object being written
     *   (array size: iNbAttPart + iNbAttObj)
     * @param ioAttVals
     *   Values of the given attributes for the part and the object being written
     *   (array size: iNbAttPart + iNbAttObj)
     * <p>
     * @return
     * The method must return 0 if the operation is accepted or error code if it is 
     * refused. In such a case, an error message is displayed in an errornotification
     * panel and the operation is canceled.<BR>
     * The attributes given in the list are the ones that were valuated in the insert
     * panel. If you wish to valuate a "hidden" attribute, you will have to reallocate
     * AttNames and AttVals and add the attribute and its value.
     */
    static Error ComputeValuesBeforeWrite(const int iUpdate,
					  const char* const iCaenv, 
					  const char* const iCatab,
					  int *iNbAttPart, 
					  int *iNbAttObj,
					  char ***ioAttNames, 
					  char ***ioAttVals);
    
    /**
      * Allows to check attribute values before updating an object.
      * <p>
      * Attributes given in the list are the ones that were valuated in the update
      * panel. If you wish to valuate a "hidden" attribute, you will have to reallocate
      * AttNames and AttVals and add the attribute and its value.
      * <p>
      * @param iobj
      *    object which is going to be updated
      * @param iONbAtt  
      *   Number of attributes given for the object
      * @param ioAttNames   
      *   Names of the given attributes
      * @param ioAttVals
      *   Values of the given attributes
      * <p>
      * @return
      *   The method must return 0 if the operation is accepted or error code if it is 
      *   refused. In such a case, an error message is displayed in an error notification
      *   panel and the operation is canceled.
      */
    static Error UpdateObject(const ObjID &iObj,
			      int *ioNbAtt,
			      char ***ioAttNames, char ***ioAttVals);


    //----------------------
    // Long Field management
    //----------------------
    /**
     * Allows to check source location or to customize LF management.
     * <p>
     * @param iCaenv
     *   Owner of the environment into which the object is being written
     * @param iCatab
     *   Name of the table into which the object is being written
     * @param iSource
     *   Where the LF (long field) comes from:<BR>
     *   <tt>1</tt>  File<BR>
     *   <tt>2</tt>  CATIA V4<BR>
     *   <tt>3</tt>  New one (empty)<BR>
     *   <tt>4</tt>  Object in database<BR>
     *   <tt>5</tt>  Document<BR>
     *   <tt>7</tt>  DENEB product
     * @param ioPath
     *   Significant only when <tt>iSource</tt> is equal to "File" (1).It is the full path
     *   of the file.
     * @param ioObj
     *   Significant only when <tt>iSource</tt> is equal to "Object in database" (4).
     * @param ioRepository
     *   Where the LF is going to be written   
     * <p>
     * @return
     * The method must return 0 if the operation is accepted or error code if it is
     * refused. In such a case, an error message is displayed in an error notification
     * panel and the operation is canceled.
     */
    static Error CheckLFBeforeWrite(const char* const iCaenv, 
				    const char* const iCatab,
				    const int iSource,
				    char **ioPath,
				    ObjID *ioObj,
				    char** ioRepository);
					  

    //----------------------------
    // Version/revision management
    //----------------------------
    /**
     * Provides a list of proposed values upon creation/update of a version. 
     * <p>
     * If number of values is greater than 0, the values will be displayed in a list
     * inside input panel.
     * <p>
     * Note: <b>used in NEW revisioning mode</b>  
     * <p>
     * @param iobj
     *    object from which we want to create a new version 
     * @param iMaturity
     *    maturity of the object
     * @param iLastRev 
     *    last existing version number
     * @param iCurRev
     *    version number of the object being handled
     * @param oNbVerValues 
     *    number of values returned
     * @param oVerValues 
     *    list of values (memory has to be allocated within the user exit; it is
     *    freed by the caller)
     * @param oAuthorized
     *    decides if output values are to be taken as authorized values (default
     *    is: help values)
     */
    static void GetNewVersionListOfValues(const ObjID& iObj         ,
                                          const char*  iMaturity    , 
                                          const char*  iLastVer     ,
                                          const char*  iCurVer      ,
                                          int*         oNbVerValues ,
                                          char***      oVerValues   ,
                                          int*         oIsAuthorized);
    
    /**
     * Provides a list of proposed values upon creation/update of a revision.
     * <p>
     * If number of values is greater than 0, the values will be displayed in a list
     * inside input panel.
     * <p>
     * Note: <b>used in NEW revisioning mode</b>
     * <p>
     * @param iObj
     *    object from which we want to create a new revision
     * @param iMaturity
     *    maturity of the object
     * @param iLastRev
     *    last existing revision number
     * @param iCurRev
     *     revision number of the object being handled
     * @param oNbRevValues
     *     number of values returned
     * @param oAuthorized
     *    decides if output values are to be taken as authorized values (default
     *    is: help values)
     */
    static void GetNewRevisionListOfValues(const ObjID& iObj        ,
                                           const char*  iMaturity   ,
                                           const char*  iLastRev    ,
                                           const char*  iCurRev     ,
                                           int*         oNbRevValues,
                                           char***      oRevValues  ,
                                           int*         oAuthorized );
  
    /**
     * Checks the validity of the keyed value for revision.
     * <p>
     * Note: <b>used in NEW revisioning mode</b>
     * <p>
     * @param iObj
     *     object from which we want to create a new version
     * @param iMaturity
     *     maturity of the object
     * @param iPrevVer
     *     previous version number of the object
     * @param iNewVer
     *     new version number
     * @param iAction
     *    create or update operation
     * <p>
     * @return
     *     The method must return an Error with Ier=0 if the operation is accepted or 
     *     error code if it is refused.<BR>
     *     In such a case, an error message is displayed in an error notification panel.
     */
    static Error IsAValidNewVersion(const ObjID&       iObj     ,
                                    const char*        iMaturity, 
				    const char*        iPrevVer ,
                                    const char*        iNewVer  ,
                                    const RevVerAction iAction  );
  
    /**
     * Checks the validity of the keyed value for revision.
     *<p>
     * Note: <b>used in NEW revisioning mode</b>
     *
     * @param iObj
     *     object from which we want to create a new revision
     * @param iMaturity
     *     maturity of the object
     * @param iPrevRev
     *     previous revision number of the object
     * @param iNewRev
     *     new revision number
     * @param iAction
     *    create or update operation
     *
     * @return
     *     The method must return an Error with Ier=0 if the operation is accepted or 
     *     error code if it is refused.<BR>
     *     In such a case, an error message is displayed in an errornotification panel.
     */
    static Error IsAValidNewRevision(const ObjID&       iObj     ,
                                     const char*        iMaturity, 
                                     const char*        iPrevRev ,
                                     const char*        iNewRev  ,
                                     const RevVerAction iAction  );
    
    /**
     * Checks the validity of the keyed value for revision duplication when versioning
     * a part.
     * <p>
     * Note: <b>used in NEW revisioning mode when declarative
     *        CATCDMA.VPM.VERSION.SELECTREVISION.INPUT is set to TRUE</b>
     * <p>
     * @param iObj
     *     object from which we want to create a new revision
     * @param iPrevRev
     *     previous revision number of the object
     * @param iNewRev
     *     new revision number
     * <p>
     * @return
     *     The method must return an Error with Ier=0 if the operation is accepted or
     *     error code if it is refused.<BR>
     *     In such a case, an error message is displayed in an error notification panel.
     */
    static Error IsAValidDuplicatedRevision(const ObjID& iObj    , 
                                            const char*  iPrevRev,
                                            const char*  iNewRev );
    
    /**
     * Checks the validity of the keyed value for version or revision.
     *<p>
     * Note: <b>used in OLD revisioning mode </b> 
     * @param iObj
     *    object from witch we want to create a new version or revision
     * @param iVersion
     *    version of the part
     * @param iOldVerRev
     *    version or revision of iobj
     * @param iNewVerRev
     *    version or revision keyed
     * <p>
     * @return
     *    The method must return an Error with Ier=0 if the operation is accepted or 
     *    error code if it is refused.<BR>
     *    In such a case, an error message is displayed in an error notification panel.
     */
    static Error IsAValidRevisionVersion(const ObjID& iObj         ,
                                         const char   iVersion[4]  ,
                                         const char   iOldVerRev[4],
                                         const char   iNewVerRev[4]);
    
    /**
     * Called before creating new part version.
     *<p>
     * Note: <b>used in BOTH revisioning mode</b>
     * <p>
     * @param iObj
     *     object from which we want to create a new version
     * @param iNewPartNumber
     *     <tt>0</tt> same part number<BR>
     *     <tt>1</tt> new part number
     */
    static Error AuthorizedNewVersion(const ObjID& iObj          ,
                                      const int    iNewPartNumber); 
    
    /**
     * Called before creating new representation revision.
     *<p>
     * Note: <b>used in BOTH revisioning mode</b>
     *<p>
     * @param iObj
     *     object from which we want to create a new revision
     */
    static Error AuthorizedNewRevision(const ObjID& iObj);
  
    /**
     * Called before duplicating onject during Massiveduplicate.
     *<p>
     * @param iObj
     *     object we want to duplicate
     * @param iTypeOfCheck
     *     type of check : 0 for check on part on which the command is called. 1 for check on intermidiate part
     * <p>
     * @return
     *   0 id Duplication is not allowed, 1 if allowed
     */
    static int IsMassiveDuplicateAllowed(const ObjID& iObj, int iTypeOfCheck);
  

    /**
     * Called after link duplicate.
     * <p>
     * @param iOld
     *     is the source link.
     * @param iNew
     *     is the newly created link.
     */
    static Error DuplicateLink(const ObjID& iOld, const ObjID& iNew);
    

     /**
     * Called after replace (but before commit).
     * <p>
     * @param iOldLink
     *     is the original link.
     * @param iNewLink
     *     is the newly created link.
     */
    static void Replace(const ObjID& iOldLink, const ObjID& iNewLink);


    /**
     * Get candidate parts when performing a "Replace / part custom...".
     * <p>
     * If provided, the output parts will be listed in VPM so the user can choose
     * with which one the initial part should be replaced with.
     * <p>
     * @param iFather
     *    Parent part object (when called from PSN graph).
     * @param iLink
     *    Link object (when called from PSN graph).
     * @param iChild
     *    Child part object to replace.
     * @param oNbCandidateChild
     *    Number of candidate parts for replace.
     * @param oCandidateChildren
     *    Candidate part objects.
     * <p>
     * @return
     *    If Error.Ier equals 0, the replace operation continues, otherwise an error
     *    message is displayed.
     */
    static Error GetReplaceCustomParts(const ObjID& iFather           ,
                                       const ObjID& iLink             ,
                                       const ObjID& iChild            ,
                                       int&         oNbCandidateChild ,
                                       ObjID*&      oCandidateChildren);

    /**
     * Get candidate revisions when asking from V5 all revisions of a specified document.
     * <p>
     * @param iTypeFather
     *    Type of search<BR>
     *    "All"   : look for all revisions of all versions of the part of "iInitialObject"<BR>
     *    "InWork": look for all revisions of the "InWork" part of "iInitialObject"
     * @param iInitialObject
     *    Object used to search
     * @param oSearchedPerformed
     *    = 1 if search has been performed<BR>
     *    = 0 if search has not been performed
     * @param oNbRevisions
     *    Number of candidate revisions
     * @param oRevisions
     *    Candidate revisions
     * <p>
     * @return
     *    If Error.Ier equals 0, processus has determined all revisions according to
     *    the specified Type
     */
    static Error GetRevisions(const char   *  iType,
                              const ObjID  &  iInitialObject,
                                    int    &  oSearchedPerformed,
                                    int    &  oNbRevisions,
                                    ObjID *&  oRevisions);

    //------------------------
    //
    //-----------------------
    /**
     * Allows user to perform specific action before object creation.
     * <p>
     * @param iCaenv
     *    Environment into which the object is going to be created
     * @param iObject
     *    Object (type) to be created
     * <p>
     * @return 
     *    If the method return an error, the creation is canceled and an error message 
     *    will be displayed in an error notification panel.
     */
    static Error BeforeCreateUserAction(const char* const iCaenv, const char* const iObject);


    //------------------------
    //
    //-----------------------
    /**
     * Allows user to perform specific action after write.
     * <p>
     * @param iObj
     *    object created
     * <p>
     * @return 
     *    If the method return an error, a rollback will be performed and an error message
     *    will be displayed in an error notification panel.
     */
    static Error AfterWriteUserAction(const ObjID& iObj);
    
    
    //------------------------
    //
    //-----------------------
    /**
     * Allows user to perform specific action after commit.
     * <p>
     * @return 
     *    If the method return an error, an error message will be displayed in an error
     *    notification panel.
     */
    static Error AfterCommitUserAction();
    
    
    //------------------------
    //
    //-----------------------
    /**
     * Allows user to perform specific action after rollback.
     */
    static void AfterRollbackUserAction();
    
    
    //---------------------
    //
    //--------------------
    /**
     * User exit called when user print result panel.
     * <p>
     * @return 
     *    an HTML color (for instance: "#FFFFFF") returned char* will be deleted by VPM
     */
    static char* GetBGColorForPrint(const char* iCatab,
				    const char* iAttribute,
				    const char* iValue);

    /**
     * Called when user select OK in New config-handler panel.<BR>
     * Allows to format the config-handler Id.
     * <p>
     * @param iObj
     *     object (product) associated to the config-handler
     * @param iCHUserId
     *     user config-handler Id
     * @param oCHNewId
     *     updated (formatted) config-handler Id
     * <p>
     * @return 
     *    If the method return an error, the creation will not be performed and an error
     *    message will be displayed in an error notification panel.
     */
    static Error FormatConfigHandlerId(const ObjID& iObj, 
				       const char* iCHUserId, 
				       char **oCHNewId);

    /**
     * Called when display option selection box.<BR>
     * Allows to change options available.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization.
     * @param iObj
     *     object (product) of current dictionary
     * @param ioNbOpt
     *     number of proposed/returned options
     * @param ioOptNameList
     *     list of options name proposed/returned
     * @param ioOptTypeList
     *     list of options type proposed
     * @param iOptDescList
     *     list of options description proposed/returned
     * <p>
     * @return 
     *    If the method return an error, the error message willbe displayed in an error
     *    notification panel.
     */
    static Error OptionListToDisplay(const char *iUser, const char *iRole, const char *iOrg, 
				     const ObjID& iObj, 
				     const int iNbOpt, int &oNbOpt, 
				     char ***ioOptNameList, char ***ioOptTypeList, char ***ioOptDescList);

    /**
     * User exit called when defining effectivities on a link.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iLink
     *     link associated to the effectivities to be defined
     * @param iNbEffParam
     *    number of rows in the parameters matrix of the proposed effectivity
     * @param iEffParamPtr
     *    list of parameters (with iNbEffParam rows) for the proposed effectivity
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error AuthorizedEffDefinition(const char *iUser, const char *iRole, const char *iOrg, 
					 const ObjID& iLink,                            
					const int iNbEffParam, EffParam *iEffParamPtr );                      
    

    /**
     * User exit called when modifying/creating/deleting an effectivity.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iLink
     *     link associated to the effectivity to modify
     * @param iOldEff
     *     old effectivity (initial effectivity to modify)<BR>
     *     NB: iOldEff.Type = "" if creating a new effectivity
     * @param iNewEff
     *     new effectivity (effectivity after modifications)<BR>
     *     NB: iNewEff.Type = "" if deleting a new effectivity
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error AuthorizedEffModification(const char *iUser, const char *iRole, const char *iOrg, 
					   const ObjID& iLink,                            
					   const EffParam& iOldEff, const EffParam& iNewEff);                      

    /**
     * User exit called when modifying/creating/deleting a proposed effectivity.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iActionName
     *     action name for the proposed effectivity
     * @param iNbEffParam
     *    number of rows in the parameters matrix of the proposed effectivity
     * @param iEffParamPtr
     *    list of parameters (with iNbEffParam rows) for the proposed effectivity
     * <p>
     * @return
     *     If method return an error, parameters must be changed or commit canceled.
     */
    static Error AuthorizedProposedEffModification(const char *iUser, const char *iRole, const char *iOrg, 
						   const char *iActionName,                            
						   const int iNbEffParam, EffParam *iEffParamPtr);

    /**
     * User exit called to control the config-handlers management.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iProduct
     *     product associated to the config-handler
     * @param iCHName
     *     name of the config-handler to control the access
     * @param iAction
     *     action to control on the config-handler (cf enum CHAction)
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error ConfigHandlerControl(const char *iUser, const char *iRole, const char *iOrg, 
                                      const ObjID&   iProduct,    
                                      const char*    iCHName, 
                                      const CHAction iAction);                      

    /**
     * User exit called to specify the authorized Visibility values and the default
     * value for config-handler creation/modification.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iProduct
     *     product associated to the config-handler
     * @param iCHName
     *     name of the config-handler to control the authorized visibility (NULL for creation)
     *
     * @param oAuthorizedVisibility
     *     All authorized values to display (default = all):<BR>
     *       CH_PUBLIC:    Everyone can use the config-handler<BR>
     *       CH_PRIVATE:   Only the owner can use the config-handler<BR>
     *       CH_DATAGROUP: Everyone with UseCH process with corresponding datagroup
     *                     can use the config-handler<p>
     *
     *       exple of use:<BR>
     *            oAuthorizedVisibility = CH_PRIVATE | CH_DATAGROUP;
     *            (to authorized these two values)<p>
     *
     * @param oDefaultVisibility
     *     Default value when creating new config-handler (default = PUBLIC if authorized)<p>
     *
     *       exple of use: 
     *            oDefaultVisibility = CH_PRIVATE;
     */
    static void ConfigHandlerVisibilityControl(const char *iUser, const char *iRole, const char *iOrg, 
					       const ObjID& iProduct,    
					       const char*  iCHName, 
					       int& oAuthorizedVisibility, int& oDefaultVisibility);

    /**
     * User exit called to control the option dictionary.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iProduct
     *     dictionary product
     * @param iOptionName
     *     name of the option to control the access
     * @param iCategoryName
     *     name of the associated type(VPM1.2)/category(VPM1.3)
     * @param iCategoryOwner
     *     owner of the category (used since VPM1.3 only)
     * @param iCategoryOrg
     *     organization of the category (used since VPM1.3 only)
     * @param iAction
     *     action to control on the dictionary (see enum @href DicoAction )
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error DictionaryControl(const char *iUser, const char *iRole, const char *iOrg, 
				   const ObjID&     iProduct,    
				   const char*      iOptionName, 
				   const char*      iCategoryName, 
				   const char*      iCategoryOwner, 
				   const char*      iCategoryOrg, 
				   const DicoAction iAction);                      

    /**
     * User exit called to control the rules in option dictionary.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iProduct
     *     dictionary product
     * @param iRuleId
     *     identifier of the rule to control the access
     * @param iRuleOwner
     *     owner of the rule
     * @param iRuleOrg
     *     organization of the rule
     * @param iAction
     *     action to control on the rules of the dictionary (see enum 
     *     @href DicoRulesAction )
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error DictionaryRulesControl(const char *iUser, const char *iRole, const char *iOrg, 
				        const ObjID&          iProduct,    
				        const char*           iRuleId, 
				        const char*           iRuleOwner, 
				        const char*           iRuleOrg, 
				        const DicoRulesAction iAction);                      

    /**
     * User exit called to control the milestones.
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * @param iProduct
     *     product from milestone
     * @param iProgramName
     *     name of the associated program
     * @param iMilestoneName
     *     name of the milestone to control the access
     * @param iMilestoneType
     *     type of the milestone to control the access (1 -&gt; date ; 2 -&gt; range)
     * @param iAction
     *     action to control on the dictionary (see enum @href MilestonesAction )
     * <p>
     * @return
     *    If Error.Ier equals 0 modification is allowed, else error message is displayed.
     */
    static Error MilestonesControl(const char *iUser, const char *iRole, const char *iOrg, 
				   const ObjID&           iProduct,    
				   const char*            iProgramName, 
				   const char*            iMilestoneName, 
				   const int              iMilestoneType, 
				   const MilestonesAction iAction);                      

    /**
     * User exit called to enable/disable each commit mode (it's possible to use context
     * management GetTaskCaller() to know the current operation to commit).
     * <p>
     * @param iUser
     *     user ID
     * @param iRole
     *     user role
     * @param iOrg
     *     user organization
     * <p>
     * @param oCommitModeToDisable
     *     Modes to disable:<BR>
     *       COMMIT_ACTION:       To validate with Action<BR>
     *       COMMIT_CFG_PARAM:    Commit with parameters<BR>
     *       COMMIT_EXISTING_EFF: Commit with existing effectivity parameters<BR>
     *       COMMIT_STEP_FILE:    Update with effectivities from STEP file<p>
     *
     *       exple of use:<BR>
     *            oCommitModeToDisable = COMMIT_CFG_PARAM | COMMIT_EXISTING_EFF;
     *            (to disable these two modes)
     * <p>
     * @return
     *    If the method return an error, the error message will be displayed in
     *    an error notification panel.
     */
    static Error CommitModeControl(const char *iUser, const char *iRole, const char *iOrg, 
				   int& oCommitModeToDisable);
    
    /**
     * User exit called when a PSN graph is opened.
     * <p>
     * @param iPart
     *     root part of the graph
     * <p>
     * @return
     *    If the method return an error, graph will not be opened.
     */
    static Error OpenGraph(const int iGraphNum, const ObjID& iPart);

    /**
      * Last user-exit called before Db deconnect.
      */
    static void ExitVPMConnectedSession();

    /**
     * User exit called when "open" is called on an object(s).
     * <p>
     * @param iObj
     *     objects onto which "open" action is performed
     */
    static Error OpenObject(const char *iViewer,const int iNbObj, ObjID *iObj);


     /**
     * User exit called when "cut" is called on a PSN node. 
     * <p>
     * @param iObjs
     *     objects onto which "cut" action is performed
     */
    static void PSNCutCB(const CATLISTP(CATVPMNode) &iObjs);

    /**
     * User exit called when "copy" is called on a PSN node. 
     * <p>
     * @param iObjs
     *     objects onto which "copy" action is performed
     */
    static void PSNCopyCB(const CATLISTP(CATVPMNode) &iObjs);


    /**
     * User exit called when "paste" is called on a PSN node. 
     * <p>
     * @param iObjs
     *     objects onto which "paste" action is performed
     * @param iLink
     *     link created by paste action
     * @param iPaste
     *     pasted object
     */
    static void PSNPasteCB(CATVPMNode *iObj, SdaiAppInstance *iLink, SdaiAppInstance *iPaste);


    /**
     * User exit called to display the backup mode panel.
     * <p>
     * @param iObj
     *     object on which action is performed
     * @param iNb
     *     size iRType
     * @param iRType
     *     value when REPRES_TYPE attribute is like 'old.backup%'
     * @param oNb
     *     size of output tables (oLnkOI and oRType)<BR>
     *     <b>NB: only the "new" constructor shoud be used for ouput 
     *         table allocation</b>
     * @param oLnkOI
     *     Link between oRType -> iRType
     * @param oRType
     *     value to display
     * <p><PRE>
     * Example:   iNb = 3         oNb = 2            <BR>
     *            iRType          oLnkOI       oRType<BR>
     *            +----+          +----+       +----+<BR>
     *           0| x  |&lt;-+  +-- 0| 2  |&lt;---- 0| z* |<BR>
     *            +----+  |  |    +----+       +----+<BR>
     *           1| y  |  +--o-- 1| 0  |&lt;---- 1| x* |<BR>
     *            +----+     |    +----+       +----+<BR>
     *           2| z  |&lt;----+                       <BR>
     *            +----+
     * </PRE><p>
     * @return
     *    not used
     */
    static int BckModR(const ObjID&  iObj  
                      ,const int     iNb
                      ,      char  **iRType
                      ,      int    *oNb
                      ,      int   **oLnkOI
                      ,      char ***oRType);

    /**
     * Execute code related to the one defined in V5 API CATBatchServices::ExecuteOnServer
     * <p>
     * To be used when Catia V5 UserExit needs to access VPM functionnality.
     * The Catia V5 command has to be defined in CATBatchServices::ExecuteOnServer
     * The command is then sent to VPM where "ExecuteCommand" is called and the 
     * result sent back to CATIA V5
     * <p>
     * @param ioLng
     *    User defined int
     * 
     * @param ioMsg
     *    User defined char*
     *
     * @return
     *    not used
     */
    static void ExecuteCommand(int &ioLng, char *& ioMsg); 

    /**
     * Allows user to Accept one MML Link between both object.
     * The MML link is defined between iFrom and iTo in the MML graph.
     * In an "Impacted By" graph: iFrom ---&gt; iTo
     * In an "Impacts On"  graph: iTo   &lt;--- iFrom
     * <p>
     * @param iFrom
     *    Object holder of the link, ie. the pointing Object
     * @param iTo
     *    Object pointed by the link.
     * <p>
     * @return
     *     If method return an error, the Accept process on link is not allowed.
     */
    static Error AuthorizedMMLAccept(const ObjID& iFrom,
				     const ObjID& iTo);
    
    /**
     * User exit called when "Duplicate" is called on an object.
     * <p>
     * @param iObj
     *     objects onto which "Duplicate" action is performed
     */
    static Error AuthorizedDuplicate(const ObjID& iObj);


    static Error UpdateConfigViewValuesForSecurity(const char* const iCaenv, 
						   const char* const iCatab,
						   int *ioNbAtt,
						   char ***ioAttNames, 
						   char ***ioAttVals,
						   int *ioModified);
    
    /**
     * User exit called before Multi Site "Export" is called on an object.
     * <p>
     * @param iObj
     *     object onto which "Export" action is performed
     * @param iTargetSite
     *     Export Target Site 
     * @param oLastTS
     *     times stamp value for last succesfull Export
     *     time stamp format YYYY-MO-DD-HH.MI.SS.NNNNNN
     *     time stamp exemple 2006-02-28-23.59.59.000000
     *     time stamp ends with '\0'
     *     time stamp maximum length is 27
     *     time stamp is compared with INFO_LF TS
     *     if LastTS > INFO_LF.TS LF 
     *        are not copied in tarfile
     *        and during Import LF are not modified
     */
    static Error GetLastSuccessfullExportTimeStamp(const ObjID& iObj,
						   const char *iTargetSite,
						   char *oLastTS);
    
    
    /**
     * User exit called when creating a zone
     * <p>
     * @return
     *    If Error.Ier equals 0 create is allowed, else error message is displayed
     */
    static Error AuthorizedCreateZone(const ObjID& iObj);

    /**
     * User exit called when deleting a zone
     * <p>
     * @return
     *    If Error.Ier equals 0 delete is allowed, else error message is displayed
     */
    static Error AuthorizedDeleteZone(const ObjID& iObj);

    /**
     * User exit called before a part gets imported by the import process
     * <p>
     * @param iObj
     *     object to be imported 
     * @param oAbortImport
     *     set to 1 to abort the import
     */
    static Error BeforePartIsImported(const ObjID& iObj, int& oAbortImport);
    
    /**
     * User exit called after a part gets imported by the import process
     * <p>
     * @param iObj
     *     object to be imported 
     */
    static Error AfterPartIsImported(const ObjID& iObj);
    
    /**
     * User exit called before a part is sent to the Replication Queue.
     * <p>
     * @param iObj
     *     object to be exported 
     * @param ioTargetSite
     *     Target Site to transfer (Max size of 10) 
     * @param ioUser
     *     User Responsible (Max size of 8)
     * @param ioOrg
     *     Org Responsible (Max size of 10)
     * @param oTransfertToSite
     *     1 if a transfert to the site is to be performed
     *     0 if the transfert to site should not be performed
     */
    static Error BeforeAddPartToQueue(const ObjID& iObj, char *ioTargetSite, char *ioUser, char *ioOrg, int &ioDoTransfertToSite); 

};
#endif
